"""
weather_forecast_alert.py

Module which leverages The National Weather Service (NWS) API (api.weather.gov) to retrieve weather forecasts and alerts based on GeoLocation (latitude, longitude).

This is a free API provided by the US government

NOTE:

    - Endpoint only support US locations (including )

"""
import json
import textwrap
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import Tuple, Union

import requests
from loguru import logger as LOGGER

from dt_tools.weather.common import ForecastType, States, Unknown, WeatherLocation
from dt_tools.console.console_helper import ConsoleHelper as ch
from dt_tools.console.console_helper import ColorFG

URL_BASE_TEMPLATE="https://api.weather.gov/points/{latitude},{longitude}"
URL_ALERT_TEMPLATE="https://api.weather.gov/alerts/active?point={latitude},{longitude}"


# =========================================================================================================    
@dataclass
class ForecastDay():

    def __init__(self, lat:float, lon:float, city:str, state: str, payload: dict):
        self.location = WeatherLocation(latitude=lat, longitude=lon)
        self.city = city
        self.state = state
        self.payload: dict = payload

    def to_string(self) -> str:
        text = f'{self.name}\n'
        text += f'  Location       : {self.city} {self.state} [{self.lat_lon}]\n'
        text += f'  Timeframe      : {self.timeframe}\n'
        text += f'  Temperature    : {self.temperature} {self.temperature_unit} {self.temperature_trend}\n'
        text += f'  Prob of Percip : {self.percipitation_pct} pct\n'
        text += f'  Wind Speed     : {self.wind_speed} {self.wind_direction}\n'
        text += f'  Icon           : {self.icon}\n'
        lines = textwrap.wrap(self.short_forecast, width=90, 
                              initial_indent='  Short Desc     : ',
                              subsequent_indent=' '*19)
        text += '\n'.join(lines) + '\n'
        # text += f'  Short Desc     : {self.short_forecast}\n'
        lines = textwrap.wrap(self.detailed_forecast, width=90, 
                              initial_indent='  Detailed Desc  : ',
                              subsequent_indent=' '*19)
        text += '\n'.join(lines) + '\n'
        # text += f'  Detailed Desc  : {self.detailed_forecast}\n'
        return text
    
    @property
    def state_full(self) -> str:
        return States.translate_state_code(self.state)
    
    @property
    def _valid_forecast_day(self) -> bool:
        if len(self.payload.keys()) > 0:
            return True
        return False
    
    @property
    def name(self) -> str:
        return self.payload.get('name', Unknown.STR)
    
    @ property
    def lat_lon(self) -> str:
        return f'{self.location.latitude}/{self.location.longitude}'
    
    @property
    def timeframe(self) -> str:
        if self._valid_forecast_day:
            tm_start = datetime.fromisoformat(self.payload.get('startTime'))
            tm_end   = datetime.fromisoformat(self.payload.get('endTime'))
            t_start = datetime.strftime(tm_start, "%a %I:%M %p")
            t_end   = datetime.strftime(tm_end, "%a %I:%M %p")
            return f'{t_start} thru {t_end}'
        return Unknown.STR
    
    @property
    def temperature(self) -> int:
        return self.payload.get('temperature', Unknown.INT)     
    @property
    def temperature_unit(self) -> str:
        return self.payload.get('temperatureUnit', Unknown.STR)
    @property
    def temperature_trend(self) -> str:
        return self.payload.get('temperatureTrend', Unknown.STR)
    @property
    def percipitation_pct(self) -> int:
        return self.payload.get('probabilityOfPrecipitation', {'value': Unknown.INT}).get('value', Unknown.INT)
    @property
    def wind_speed(self) -> str:
        return self.payload.get('windSpeed', Unknown.STR)
    @property
    def wind_direction(self) -> str:
        return self.payload.get('windDirection', Unknown.STR)
    @property
    def icon(self) -> str:
        return self.payload.get('icon', Unknown.STR)
    @property
    def short_forecast(self) -> str:
        return self.payload.get('shortForecast', Unknown.STR)
    @property
    def detailed_forecast(self) -> str:
        return self.payload.get('detailedForecast', Unknown.STR)
    


# =========================================================================================================    
class AbstractEndpoint(ABC):
    def __init__(self, lat: float, lon: float, friendly_name: str = ''):
        self.location = WeatherLocation(latitude=lat, longitude=lon)
        self._friendly_name: str = friendly_name
        self._city: str = None
        self._state: str = None
        self._json_error: dict = {}

        self._last_update: datetime = None
        self._valid_payload: bool = False
        
    @property
    def latitude(self) -> float:
        return float(self.location.latitude)
    
    @property
    def longitude(self) -> float:
        return float(self.location.longitude)
    
    @property
    def name(self):
        '''Friendly name (if exists) or location'''
        if self._friendly_name == '':
            return self.location
        return self._friendly_name
    
    @property
    def city(self) -> str:
        return self._city
    
    @property
    def state(self) -> str:
        # if self._state is None:
        return self._state
        
    @property
    def state_full(self) -> str:
        return self._translate_state(self._state)
    
    @property
    def city_state(self):
        if self._city is None and self._state is None:
            return None
        return f'{self._city},{self._state}'
    
    def _translate_state(self, state: str) -> str:
        return States.translate_state_code(state)
    
    def refresh_if_needed(self, force: bool = False) -> bool:
        '''Return true if refresh needed and successful else false'''
        if self._last_update is None or (datetime.now() - self._last_update).total_seconds() > 600.0:
            self._valid_payload = self._refresh()
            if self._valid_payload:
                self._last_update = datetime.now()
            return self._valid_payload
        return False
    
    @classmethod
    def _refresh(self) -> bool:
        LOGGER.error('Must implement _refresh() function in concrete class.')
        raise NotImplementedError('_refresh()')

    def _call_endpoint(self, URL: str) -> Tuple[int, dict]:
        LOGGER.debug(f'Calling: {URL}')
        resp = requests.get(URL)
        LOGGER.debug(f'  returns: {resp.status_code}')
        if resp.status_code != 200:
            self._json_error = resp.json()

        return resp.status_code, resp.json()
    

# =========================================================================================================    
class Forecast(AbstractEndpoint):
    """
    Class to interface for weather Forecast

    Populate ForecastDay class by calling one of: 
    - forecast_for_future_day()
    - forecast_for_today()
    - forecast_for_tomorrow()
    
    Example:

        weather_forecast = Forcast(30.0694,-81.5515) # Fruit Cove, Florida
        today: ForecastDay = weather_forecast.forecast_for_today(ForecastType.DAY)
        print(today.short_forecast)

    """
    def __init__(self, lat: float, lon: float, friendly_name: str = '', base_only: bool = False):
        super().__init__(lat, lon, friendly_name)
        self._url = URL_BASE_TEMPLATE.replace('{latitude}', str(lat)).replace('{longitude}', str(lon))

        self._json_base: dict = {}
        self._json_daily_forecast:   dict = {}
        self._json_hourly_forecast: dict = {}
        if self._refresh(base_only):
            self._valid_payload = True
            self._last_update = datetime.now()

            
    def _refresh(self, base_only: bool = False) -> bool:
        self._json_error = {}
        rc, payload =  self._call_endpoint(self._url)
        if rc == 200:
            self._json_base = payload
            self._json_daily_forecast = {}
            self._json_hourly_forecast = {}
            if self._city is None:
                self._city = self._json_base['properties']['relativeLocation']['properties']['city']
                self._state = self._json_base['properties']['relativeLocation']['properties']['state']
            if not base_only:
                forecast_url = payload['properties']['forecast']
                hourly_forecast_url = payload['properties']['forecastHourly']
                rc, payload = self._call_endpoint(forecast_url)
                if rc == 200:
                    self._json_daily_forecast = payload
                    rc, payload = self._call_endpoint(hourly_forecast_url)
                    if rc == 200:
                        self._json_hourly_forecast = payload

        return (rc == 200)
    
    def forecast_for_future_day(self, days_in_future: int, time_of_day: ForecastType = ForecastType.DAY) -> Union[ForecastDay, None]:
        """
        Get weather forecast for a future day/night

        Args:
            days_in_future (int): Number of days in future (0=Today)
            time_of_day (ForecastType, optional): Forecast.DAY or Forecast.NIGHT. Defaults to ForecastType.DAY.

        Returns:
            Union[ForecastDay, None]: Forecast for target day if found, else None
        """
        # NOTE: May be bug here if current time of day has passed day/night boundary
        #       Possible that entry 0 is night (not day)
        if self._valid_payload:
            idx = (days_in_future * 2) + time_of_day.value
            payload = self._json_daily_forecast['properties']['periods'][idx]
            return ForecastDay(self.latitude, self.longitude, self.city, self.state, payload)
        LOGGER.warning(f'ForecastFutureDay({days_in_future}, {time_of_day.name}) - Invalid payload!')
        return None
    
    def forecast_for_today(self, time_of_day: ForecastType = ForecastType.DAY) -> ForecastDay:
        """
        Get weather forecast for today (or tonight)

        Args:
            time_of_day (ForecastType, optional): Forecast.DAY or Forecast.NIGHT. Defaults to ForecastType.DAY.

        Returns:
            Union[ForecastDay, None]: Forecast for today/night if found, else None
        """
        return self.forecast_for_future_day(0, time_of_day)
    
    def forecast_for_tomorrow(self, time_of_day: ForecastType = ForecastType.DAY) -> ForecastDay:
        """
        Get weather forecast for tomorrow

        Args:
            time_of_day (ForecastType, optional): Forecast.DAY or Forecast.NIGHT. Defaults to ForecastType.DAY.

        Returns:
            Union[ForecastDay, None]: Forecast for tomorrow (day or night) if found, else None
        """
        return self.forecast_for_future_day(1, time_of_day)


# =========================================================================================================    
class LocationAlerts(AbstractEndpoint):
    def __init__(self, lat: float, lon: float, friendly_name: str = '', weather: Forecast = None):
        super().__init__(lat, lon, friendly_name)

        self.loc_id = f'{lat}|{lon}'
        self._weather: Forecast = weather if weather is not None else Forecast(lat, lon, base_only=True)
        self._url = URL_ALERT_TEMPLATE.replace('{latitude}', str(lat)).replace('{longitude}', str(lon))
        self._json_alert = {}
        self.refresh_if_needed(force=True)
        self._city = self._weather.city
        self._state = self._weather.state

        # self._speak_accent: ACCENT = ACCENT.UnitedStates

    def _refresh(self) -> bool:
        self._json_error = {}
        rc, payload =  self._call_endpoint(self._url)
        if rc == 200:
            self._json_alert = payload
    
        return (rc == 200)
    
    @property
    def alert_count(self) -> int:
        return len(self._json_alert.get('features',''))
    
    def alert_id(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'id')
    def effective(self, alert_num: int) -> datetime:
        return self._get_property(alert_num, 'effective')
    def expires(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'expires')
    def status(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'status')
    def message_type(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'messageType')
    def serverity(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'severity')
    def certainty(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'certainty')
    def urgency(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'urgency')
    def event(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'event')
    def headline(self, alert_num: int) -> str:
        token = self._get_property(alert_num, 'headline').split(' by ') 
        return token[0]
    def description(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'description')
    def instruction(self, alert_num: int) -> str:
        return self._get_property(alert_num, 'instruction')

    def get_alert_idx(self, id: str) -> int:
        idx = -1
        for cnt in range(0, len(self._json_alert['features'])):
            if id == self.alert_id(cnt):
                idx = cnt
                break
    
        return idx
    
    def to_string(self) -> str:
        output = 'ALERTS-\n'
        location = f'{self.city} {self.state} [{self.location.latitude}/{self.location.longitude}]'
        location = ch.cwrap(location, fg=ColorFG.GREEN2)
        output += f'num_alerts  : {self.alert_count}\n'
        # output += f'location    : {self.city} {self.state} [{self.location.latitude}/{self.location.longitude}]\n'
        output += f'location    : {location}\n'
        cnt = 0
        for cnt in range(0, self.alert_count):
            output += f'[{cnt+1}] Alert {self.alert_id(cnt)}\n'
            # output += f'    refid       : {self.reference_id(cnt)}\n'
            output += f'    effective   : {self.effective(cnt)} - {self._convert_date(self.effective(cnt), True)}\n'
            output += f'    expires     : {self.expires(cnt)} - {self._convert_date(self.expires(cnt))}\n'
            output += f'    type        : {self.message_type(cnt)}\n'
            output += f'    severity    : {self.serverity(cnt)}\n'
            output += f'    certainty   : {self.certainty(cnt)}\n'
            output += f'    urgency     : {self.urgency(cnt)}\n'
            output += f'    event       : {self.event(cnt)}\n'
            output += f'    headline    : {self.headline(cnt)}\n'
            if len(self.description(cnt)) < 80:
                output += f'    description : {self.description(cnt)}\n'
            else:
                output +=  '    description :\n'
                lines = textwrap.wrap(self.description(cnt), width=120, 
                                    initial_indent='       ',
                                    subsequent_indent=' '*7)
                output += '\n'.join(lines) + '\n'
            if len(self.instruction(cnt)) < 80:
                output += f'    instruction : {self.instruction(cnt)}\n'
            else:
                output +=  '    instruction :\n'
                lines = textwrap.wrap(self.instruction(cnt), width=120, 
                                    initial_indent='       ',
                                    subsequent_indent=' '*7)
                output += '\n'.join(lines) + '\n'

            output += '\n'
        return output
        
    def _convert_date(self, str_iso_date: str, inc_month: bool = False) -> str:
        token = str_iso_date
        if token is not None and token != Unknown.STR:
            date_val = datetime.fromisoformat(token) 
            date_fmt = "%B %d at %I:%M%p" if inc_month else "%I:%M%p"
            ret_date = date_val.strftime(date_fmt)
        return ret_date

    def _get_property(self, alert_num: int, key: str) -> str:
        val = Unknown.STR
        if alert_num < self.alert_count:
            props: dict = self._json_alert['features'][alert_num]['properties']
            val = props.get(key, Unknown.STR)
            if val is None:
                val = Unknown.STR
        return val
    


def main():
    lat, lon = (30.0694,-81.5515)       # Cimarrone
    # lat, lon = (40.6561391,-74.0035868)   # Brooklyn
    # lat, lon = (45.5537619,-122.7071236)  # Portland 
    weather_forecast = Forecast(lat, lon)
    # print(json.dumps(weather._json_base, indent=2))
    LOGGER.warning(f'Forecast for: {weather_forecast.city_state}')    
    today: ForecastDay = weather_forecast.forecast_for_today(ForecastType.DAY)
    if today is not None:
        LOGGER.debug(f'\n{today.to_string()}')
        LOGGER.success('today short forecast:')
        LOGGER.info(f'  {today.short_forecast}')
        LOGGER.success('today long forecast:')
        LOGGER.info(f'  {today.detailed_forecast}')
        LOGGER.info('')
    tonight: ForecastDay = weather_forecast.forecast_for_today(ForecastType.NIGHT)
    if tonight is not None:
        LOGGER.debug(f'\n{json.dumps(weather_forecast._json_base)}')
        LOGGER.success('tonight short forecast:')
        LOGGER.info(f'  {tonight.short_forecast}')
        LOGGER.success('tonight long forecast:')
        LOGGER.info(f'  {tonight.detailed_forecast}')
        LOGGER.info('')

    alert = LocationAlerts(lat, lon)
    LOGGER.info(f'\n{alert.to_string()}')
    for i in range(0, alert.alert_count):
        LOGGER.info(alert.headline(i))
        LOGGER.info(f'  {alert.description(i)}')
        LOGGER.info(f'  {alert.instruction(i)}')

if __name__ == "__main__":
    import dt_tools.logger.logging_helper as lh
    lh.configure_logger(log_level="INFO", log_format=lh.DEFAULT_DEBUG_LOGFMT, brightness=False)
    main()



"""
Get sunrise and sunset for specific GPS coordinate

Example:

    sun = Sun()
    print('Sunrise: {sun.sunrise}  Sunset: {sun.sunset}')
"""
import calendar
import datetime
import math

from dateutil import tz
from loguru import logger as LOGGER
from zoneinfo import ZoneInfo

from dt_tools.geoloc.geoloc import GeoLocation

GEO = GeoLocation()

class SunTimeException(Exception):
    def __init__(self, message):
        super(SunTimeException, self).__init__(message)

class Sun:
    """
    Approximated calculation of sunrise and sunset datetimes. Adapted from:
    https://stackoverflow.com/questions/19615350/calculate-sunrise-and-sunset-times-for-a-given-gps-coordinate-within-postgresql
    """
    def __init__(self, lat, lon):
        """
        Sun object located at lat, lon

        Args:
            lat (_type_): latitude
            lon (_type_): longitude
        """
        self._lat = float(lat)
        self._lon = float(lon)
        LOGGER.debug(f'Sun({lat},{lon})')

    def time_now_at(self) -> datetime.datetime:
        """
        Current local time at GPS Coordinates

        Returns:
            datetime.datetime: current date time in GPS coordinate local time.
        """
        # geo = GeoLocation()
        GEO.get_location_via_lat_lon(self._lat, self._lon)
        utc_now = datetime.datetime.now(datetime.timezone.utc)
        gps_now = utc_now.astimezone(ZoneInfo(GEO.tz_name))
        return gps_now

    def get_sunrise_time(self, date:datetime.date=None) -> datetime.datetime:
        """
        Get sunrise time at GPS coordinates in UTC format

        Args:
            date (datetime.date, optional): Target date, if none, today. Defaults to None.

        Raises:
            SunTimeException: When the sun does not rise at this location on the specified date.

        Returns:
            _type_: UTC datetime
        """
        date = datetime.date.today() if date is None else date
        sr = self._calc_sun_time(date, True)
        if sr is None:
            raise SunTimeException(f"The sun never rises on this location (on the specified date) | {date} {self._lat}/{self._lon}")
        
        return sr

    def get_local_sunrise_time(self, date:datetime.date=None, local_time_zone=tz.tzlocal()) -> datetime.datetime:
        """
        Get sunrise time in local timezone format

        Args:
            date (datetime.date, optional): Target date, if none, today. Defaults to None.

        Raises:
            SunTimeException: When the sun does not rise at this location on the specified date.

        Returns:
            _type_: datetime - Local timezone datetime
        """
        date = datetime.date.today() if date is None else date
        sr = self._calc_sun_time(date, True)
        if sr is None:
            raise SunTimeException("The sun never rises on this location (on the specified date)")
        
        return sr.astimezone(local_time_zone)

    def get_sunset_time(self, date=None) -> datetime.datetime:
        """
        Get sunset time at GPS coordinates in UTC format

        Args:
            date (datetime.date, optional): Target date, if none, today. Defaults to None.

        Raises:
            SunTimeException: When the sun does not set at this location on the specified date.

        Returns:
            _type_: UTC datetime
        """
        date = datetime.date.today() if date is None else date
        ss = self._calc_sun_time(date, False)
        if ss is None:
            raise SunTimeException(f"The sun never sets on this location (on the specified date) | {date} {self._lat}/{self._lon}")
        
        return ss

    def get_local_sunset_time(self, date=None, local_time_zone=tz.tzlocal()) -> datetime.datetime:
        """
        Get sunset time in local timezone format

        Args:
            date (datetime.date, optional): Target date, if none, today. Defaults to None.

        Raises:
            SunTimeException: When the sun does not set at this location on the specified date.

        Returns:
            _type_: datetime - Local timezone datetime
        """
        date = datetime.date.today() if date is None else date
        ss = self._calc_sun_time(date, False)
        if ss is None:
            raise SunTimeException(f"The sun never sets on this location (on the specified date) | {date} {self._lat}/{self._lon}")
        
        return ss.astimezone(local_time_zone)

    def get_gps_sunrise(self, date=None) -> datetime.datetime:
        utc_sunrise = self.get_sunrise_time(date)
        LOGGER.trace(f'UTC: {utc_sunrise}')
        GEO.get_location_via_lat_lon(self._lat, self._lon)

        gps_sunrise = utc_sunrise.astimezone(ZoneInfo(GEO.tz_name))
        LOGGER.trace(f'GPS: {gps_sunrise} {GEO.tz_name}')
        return gps_sunrise
    
    def get_gps_sunset(self, date=None) -> datetime.datetime:
        utc_sunset = self.get_sunset_time(date)
        gps_sunset = utc_sunset.astimezone(ZoneInfo(GEO.tz_name))
        return gps_sunset


    def _calc_sun_time(self, date, isRiseTime=True, zenith=90.8) -> datetime.datetime:
        """
        Calculate sunrise or sunset date.
        :param date: Reference date
        :param isRiseTime: True if you want to calculate sunrise time.
        :param zenith: Sun reference zenith
        :return: UTC sunset or sunrise datetime
        :raises: SunTimeException when there is no sunrise and sunset on given location and date
        """
        # isRiseTime == False, returns sunsetTime
        day = date.day
        month = date.month
        year = date.year

        TO_RAD = math.pi / 180.0

        # 1. first calculate the day of the year
        N1 = math.floor(275 * month / 9)
        N2 = math.floor((month + 9) / 12)
        N3 = 1 + math.floor((year - 4 * math.floor(year / 4) + 2) / 3)
        day_of_the_year = N1 - (N2 * N3) + day - 30

        # 2. convert the longitude to hour value and calculate an approximate time
        longitude_hour = self._lon / 15

        if isRiseTime:
            t = day_of_the_year + ((6 - longitude_hour) / 24)
        else:  # sunset
            t = day_of_the_year + ((18 - longitude_hour) / 24)

        # 3. calculate the Sun's mean anomaly
        mean_anomaly = (0.9856 * t) - 3.289

        # 4. calculate the Sun's true longitude
        true_longitude = mean_anomaly + \
                        (1.916 * math.sin(TO_RAD * mean_anomaly)) + \
                        (0.020 * math.sin(TO_RAD * 2 * mean_anomaly)) + 282.634
        true_longitude = self._force_range(true_longitude, 360)  # NOTE: L adjusted into the range [0,360)

        # 5a. calculate the Sun's right ascension
        right_ascension = (1 / TO_RAD) * math.atan(0.91764 * math.tan(TO_RAD * true_longitude))
        right_ascension = self._force_range(right_ascension, 360)  # NOTE: RA adjusted into the range [0,360)

        # 5b. right ascension value needs to be in the same quadrant as L
        l_quadrant = (math.floor(true_longitude / 90)) * 90
        right_ascension_quadrant = (math.floor(right_ascension / 90)) * 90
        right_ascension += (l_quadrant - right_ascension_quadrant)

        # 5c. right ascension value needs to be converted into hours
        right_ascension = right_ascension / 15

        # 6. calculate the Sun's declination
        sinDec = 0.39782 * math.sin(TO_RAD * true_longitude)
        cosDec = math.cos(math.asin(sinDec))

        # 7a. calculate the Sun's local hour angle
        cosH = (math.cos(TO_RAD * zenith) - (sinDec * math.sin(TO_RAD * self._lat))) / (
            cosDec * math.cos(TO_RAD * self._lat)
        )
        if cosH > 1:
            return None  # The sun never rises on this location (on the specified date)
        if cosH < -1:
            return None  # The sun never sets on this location (on the specified date)

        # 7b. finish calculating H and convert into hours
        if isRiseTime:
            hours = 360 - (1 / TO_RAD) * math.acos(cosH)
        else:  # setting
            hours = (1 / TO_RAD) * math.acos(cosH)

        hours = hours / 15

        # 8. calculate local mean time of rising/setting
        mean_time = hours + right_ascension - (0.06571 * t) - 6.622

        # 9. adjust back to UTC
        utc_time = mean_time - longitude_hour
        utc_time = self._force_range(utc_time, 24)  # UTC time in decimal format (e.g. 23.23)

        # 10. Return
        hr = self._force_range(int(utc_time), 24)
        min = round((utc_time - int(utc_time)) * 60, 0)
        if min == 60:
            hr += 1
            min = 0

        # 10. check corner case https://github.com/SatAgro/suntime/issues/1
        if hr == 24:
            hr = 0
            day += 1

            if day > calendar.monthrange(year, month)[1]:
                day = 1
                month += 1

                if month > 12:
                    month = 1
                    year += 1

        return datetime.datetime(year, month, day, hr, int(min), tzinfo=tz.tzutc())

    @staticmethod
    def _force_range(v, max):
        # force v to be >= 0 and < max
        if v < 0:
            return v + max
        elif v >= max:
            return v - max

        return v


def __display_sun_data(lat, lon, show_americas: bool = False):
    datetime_format="%a %m/%d %I:%M %p"

    # geo = GeoLocation()
    GEO.get_location_via_lat_lon(lat=lat,lon=lon)
    LOGGER.debug('== geo ===============================')
    for line in GEO.to_string().splitlines():
        LOGGER.debug(f'  {line}')

    sun = Sun(lat, lon)
    
    LOGGER.warning(f'Sunrise/Sunset for {GEO.display_name}')
    utc_sunrise = sun.get_sunrise_time()
    utc_sunset  = sun.get_sunset_time()

    local_sunrise = sun.get_local_sunrise_time()
    local_sunset  = sun.get_local_sunset_time()

    gps_sunrise = sun.get_gps_sunrise()
    gps_sunset  = sun.get_gps_sunset()

    LOGGER.info(f'  current time : {datetime.datetime.strftime(sun.time_now_at(), datetime_format)}  [{GEO.tz_name}]')
    LOGGER.info(f'      latitude : {lat}  longitude: {lon} ')
    LOGGER.info('')
    LOGGER.info('Sunrise                  Sunset                   Times at location')
    LOGGER.info('------------------------ ------------------------ ------------------------------------------')
    try:
        LOGGER.info(f"{local_sunrise.strftime(datetime_format):24} {local_sunset.strftime(datetime_format):24} Local [{local_sunrise.tzname()}]")
        LOGGER.info(f"{utc_sunrise.strftime(datetime_format):24} {utc_sunset.strftime(datetime_format):24} UTC")
        LOGGER.info(f"{gps_sunrise.strftime(datetime_format):24} {gps_sunset.strftime(datetime_format):24} {GEO.display_name}")

    except SunTimeException as e:
        print("Error: {0}".format(e))

    LOGGER.info('')

if __name__ == "__main__":
    """
    https://geocoding.geo.census.gov/geocoder/Geocoding_Services_API.pdf
    https://geocoding.geo.census.gov/geocoder/locations/address?street=4833+nahane+way&zip=32259&benchmark=Public_AR_Current&format=html

    returntype=locations|geographies
    searchtype=onlineaddress|address|coordinates
    benchmark=Public_AR_Current
    address= or street  state zip
    format=json|html

    """

    import dt_tools.logger.logging_helper as lh

    lh.configure_logger(log_level="INFO", log_format=lh.DEFAULT_DEBUG_LOGFMT, brightness=False)
    geo = GeoLocation()

    # gps = (30.069344, -81.551315)
    # LOGGER.info(f'By Lat/Lon: {gps[0]}, {gps[1]}')
    # __display_sun_data(gps[0], gps[1])
    # LOGGER.info('')

    addr = '511 East Magnolia, Bellingham, WA'
    LOGGER.info(f'By Address: {addr}')
    if GEO.get_location_via_address_string(addr):
        __display_sun_data(GEO.lat, GEO.lon)
    else:
        LOGGER.warning('Invalid address...')
    LOGGER.info('')

    addr = 'West 3rd Street, Duluth, MN'
    LOGGER.info(f'By Address: {addr}')
    if GEO.get_location_via_address_string(addr):
        __display_sun_data(GEO.lat, GEO.lon)
    else:
        LOGGER.warning('Invalid address...')
    LOGGER.info('')

    # zip=95356
    # LOGGER.info(f'By zip: {zip}')
    # if GEO.get_location_via_zip(zip, 'US'):
    #     __display_sun_data(GEO.lat, GEO.lon)
    # else:
    #     LOGGER.warning('Invalid zip...')
    # LOGGER.info('')
    



"""
This module demonstrates the Geolocation module functionality.


"""

from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper
from dt_tools.geoloc.census_geoloc import Census_GeoLocation as CensusGeoLocation
from dt_tools.geoloc.census_geoloc import GeoLocationAddress
from dt_tools.geoloc.geoloc import GeoLocation


def _print_object(obj):
    for line in obj.to_string().splitlines():
        LOGGER.info(f'  {line}')
    payload = getattr(obj,'_json_payload')
    if payload is not None:
        LOGGER.info('  Payload-')
        for k, v in payload.items():
            LOGGER.info(f'    {k:15} {v}')

def _demo_census_geoloc():
    LOGGER.info('')
    LOGGER.info('Census Geolocation.')
    LOGGER.info('  Must supply street, city, state OR street, zip code')
    LOGGER.info('')

    geo = CensusGeoLocation()
    street = '87 Rockefeller Center'
    city = 'New York'
    state = 'NY'
    zip = None
    locations = geo.lookup_address('87 Rockefeller Center', 'New York', 'NY')
    loc: GeoLocationAddress = None
    LOGGER.info(f'Lookup  - street: {street}  city: {city}  state: {state}  zip: {zip}')
    for loc in locations:
        # LOGGER.success(f'Returns - lat: {loc.latitude:.4f}  lon: {loc.longitude:.4f}  address: {loc.address}')
        _print_object(loc)
    LOGGER.info('')

def _demo_geoloc():
    LOGGER.info('')
    LOGGER.warning('Geolocation (geocode.maps.co)')
    LOGGER.info('')

    geo = GeoLocation()
    LOGGER.warning('Retrieve via address string')
    address = 'Metlife Stadium, NJ'
    geo.get_location_via_address_string(address)
    LOGGER.info(f'  {address}')
    LOGGER.info('Returns -')
    _print_object(geo)
    LOGGER.info('')

    LOGGER.warning('Retrieve via address')
    address = '511 East Magnolia, Bellingham, WA 98225'
    geo.get_location_via_address(city='Bellingham', state='WA', house='511', street='East Magnolia', zip=98225)
    LOGGER.info("  city='Bellingham', state='WA', house='511', street='East Magnolia', zip=98225")
    LOGGER.info('Returns -')
    _print_object(geo)
    LOGGER.info('')

    LOGGER.warning('Retrieve via latitude/longitude')
    lat = geo.lat
    lon = geo.lon
    geo.get_location_via_lat_lon(lat, lon)
    LOGGER.info(f'  lat: {lat}  lon: {lon}')
    LOGGER.info('Returns -')
    _print_object(geo)
    LOGGER.info('')

    LOGGER.warning('Retrieve via IP')
    geo.get_location_via_ip()
    LOGGER.info(f'  IP resolved as {geo.ip}')
    LOGGER.info('Returns -')
    # for line in geo.to_string().splitlines():
    #     LOGGER.info(f'  {line}')
    _print_object(geo)
    LOGGER.info('')

def demo():
    OSHelper.enable_ctrl_c_handler()
    LOGGER.info('-'*80)
    LOGGER.info('dt_misc_geoloc demo')
    LOGGER.info('-'*80)

    _demo_census_geoloc()
    _demo_geoloc()

    LOGGER.info('')
    LOGGER.info('Demo complete.')

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level='INFO', brightness=False)
    demo()



"""
This module will execute the dt_misc package demonstrations, which include:

- GeoLocation
- Weather (current, forecast, alerts)
- Sound (TTS - text to speech)

To Run:
    ``poetry run python -m dt_tools.cli.demos.dt_misc_demo``

"""
from loguru import logger as LOGGER
import dt_tools.cli.demos.dt_geoloc_demo as geoloc_demo
import dt_tools.cli.demos.dt_weather_demo as weather_demo
import dt_tools.cli.demos.dt_misc_sound_demo as sound_demo
import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper
from dt_tools.os.project_helper import ProjectHelper

def demo():
    DEMOS = {
        "GeoLocation Demo": geoloc_demo,
        "Weather demo": weather_demo,
        "Sound demo": sound_demo
    }
    l_handle = lh.configure_logger(log_level="INFO", brightness=False)  # noqa: F841
    LOGGER.info('='*80)
    version = f'v{ProjectHelper.determine_version("dt-misc")}'
    LOGGER.info(f'dt_misc_demo {version}', 80)
    LOGGER.info('='*80)
    LOGGER.info('')
    for name, demo_module in DEMOS.items():
        if input(f'Run {name} (y/n)? ').lower() == 'y':
            demo_module.demo()  
            LOGGER.info('') 

if __name__ == '__main__':
    OSHelper.enable_ctrl_c_handler()
    demo()
                                                      
    LOGGER.success("That's all folks!!")


"""
This module demonstrates the Sound class functionality.

- Speak with accents.
- Vary speech speed/cadence.
- Speak the contents of a text file.

"""
import pathlib
from datetime import datetime as dt

from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper
from dt_tools.os.project_helper import ProjectHelper
from dt_tools.sound.helper import Accent, Sound
import requests

DEMO_SPEED = 1.25

def _get_quote_of_the_day() -> str:
    url = "https://zenquotes.io/api/random"
    text = ''
    try:
        resp = requests.get(url)
        if resp.status_code == 200:
            resp_json = resp.json()
            quote = resp_json[0]['q']
            author = resp_json[0]['a']
            text = f'{quote}...\n  {author}...'
    except Exception as ex:
        LOGGER.error(repr(ex))
        text = ''
    return text

def _accent_demo():
    Sound().speak('Lets start out with an Accent demonstration.', speed=DEMO_SPEED)
    LOGGER.info('')
    LOGGER.info('Accents demo')
    LOGGER.info('------------')
    LOGGER.info('  Code')
    LOGGER.info('  ----------------------------------------------------------------------------------------------------')
    for accent in Accent:
        LOGGER.success(f'  Sound().speak(f"This is a test, with a {accent.name} accent.", accent={accent}, speed={DEMO_SPEED})')
        Sound().speak(f'This is a test, with a {accent.name} accent.', accent=accent, speed=DEMO_SPEED, wait=True)
    
    input('\nPress Enter to continue... ')

def _speed_demo():
    Sound().speak('Lets continue with a speach speed demonstration.', speed=DEMO_SPEED)
    LOGGER.info('')
    LOGGER.info('Speech speed (cadence) demo')
    LOGGER.info('----------------------------')
    LOGGER.info('  Code')
    LOGGER.info('  ----------------------------------------------------------------------------------------------------')
    for speed in [0.75, 1.0, 1.25, 1.5, 2.0]:
        LOGGER.success(f'  Sound().speak("This is a test with the speed set to {speed}.", speed={speed})')
        Sound().speak(f'This is a test, with the speed set to {speed}.', speed=speed, wait=True)
    
    input('\nPress Enter to continue... ')

def _file_demo():
    LOGGER.info('')
    LOGGER.info('Text file sound demo')
    LOGGER.info('--------------------')
    snd = Sound()
    snd.speak('This demo will create a text file and translate it into speech.', speed=DEMO_SPEED, wait=True)
    test_file = pathlib.Path(OSHelper().get_temp_filename(prefix='dt-', dotted_suffix='.txt', target_dir='.'))
    LOGGER.info(f'- Generate text file [{test_file}].')
    date = dt.strftime(dt.now(), "%A %B %d")
    time = dt.strftime(dt.now(), "%I:%M %p")
    text = f"Today is {date} at {time}.\n"
    text += f"The operating system is {OSHelper.os_version()}.\n"
    if OSHelper.is_god():
        text += f"The current user: {OSHelper.current_user()}, is running with elevated privileges.\n"
    else:
        text += f"The current user: {OSHelper.current_user()}, is NOT running with elevated privileges.\n"
    ver, src = ProjectHelper.determine_version('dt-misc', identify_src=True)
    text += f"Version {ver} of this demo is running.\n"
    text += f"The version was determined by {src}\n"
    quote = _get_quote_of_the_day()
    if len(quote) > 0:
        text += "Now for the quote of the day:\n"
        text += f'{quote}\n'

    LOGGER.info(f'- Writing to {test_file}...')
    test_file.write_text(text)
    LOGGER.info('- Contents of file:')
    for line in text.splitlines():
        print(f'    {line}')
    LOGGER.info('')
    LOGGER.info('- Reading and speaking the text in the file.')
    LOGGER.success(f'  Sound().speak("{test_file}", speed{DEMO_SPEED})')
    snd.speak(test_file, speed=DEMO_SPEED, wait=True)
    
    test_file.unlink()

def demo():
    OSHelper.enable_ctrl_c_handler()
    LOGGER.info('-'*80)
    LOGGER.info('dt_misc_sound_helper_demo')
    LOGGER.info('-'*80)

    _accent_demo()
    _speed_demo()
    _file_demo()

    LOGGER.info('')
    salutation = 'This concludes the sound demo...  Thanks for listening.'
    Sound().speak(salutation, speed=DEMO_SPEED, wait=False)
    LOGGER.info(salutation)

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, brightness=False)
    demo()



"""
Common weather constructs

"""
from dataclasses import dataclass
from enum import Enum

from loguru import logger as LOGGER


# =========================================================================================================    
"""
Air Quality Index
"""
AQI_DESC = {
    -1: 'Unknown',
    1: 'Good',
    2: 'Moderate',
    3: 'Degraded',
    4: 'Unhealthy',
    5: 'Very Unhealthy',
    6: 'Hazardous',
}


# =========================================================================================================    
class Unknown():
    STR: str = "Unknown"
    INT: int = -1

# =========================================================================================================    
class ForecastType(Enum):
    DAY: int = 0
    NIGHT: int = 1

# =========================================================================================================    
class WeatherSymbols(Enum):
    degree = chr(176)
    N = "North"
    S = "South"
    E = "East"
    W = "West"

    def translate_compass_point(wind_dir: str) -> str:
        """
        Given the abbreviated wind direction, return full text


        Args:
            wind_dir (str): Wind direction (ie. N, S, E, W, NE, NW,...)

        Returns:
            str: Full text representation of direction (i.e. NW = North West)

        """
        resp = ''
        for char in wind_dir:
            resp += f' {WeatherSymbols[char].value}'
        return resp.lstrip()

# =========================================================================================================    
class States(Enum):
    """
    States enumeration (US and Territories)

    """
    AK = "Alaska"
    AL = "Alabama"
    AR = "Arkansas"
    AS = "American Somoa"
    AZ = "Arizonz"
    CA = "California"
    CO = "Colorado"
    CT = "Connecticut"
    DC = "District of Columbia"
    DE = "Delaware"
    FL = "Florida"
    GA = "Georgia"
    GU = "Guam"
    HI = "Hawaii"
    IA = "Idaho"
    ID = "Idaho"
    IL = "Illinois"
    IN = "Indiana"
    KS = "Kansas"
    KY = "Kentuky"
    LA = "Louisiana"
    MA = "Massachusetts"
    MD = "Maryland"
    ME = "Maine"
    MI = "Michigam"
    MN = "Minnesota"
    MO = "Missouri"
    MP = "Northern Marina Islands"
    MS = "Mississippi"
    MT = "Montana"
    NC = "North Carolina"
    ND = "North Dakota"
    NE = "Nebraska"
    NH = "New Hampshire"
    NJ = "New Jersey"
    NM = "New Mexico"
    NV = "Nevada"
    NY = "New York"
    OH = "Ohio"
    OK = "Oklahoma"
    OR = "Oregon"
    PA = "Pennsylvania"
    PR = "Puerto Rico"
    RI = "Rhode Island"
    SC = "South Carolina"
    SD = "South Dakota"
    TN = "Tenessee"
    TT = "Trust Territories"
    TX = "Texas"
    UT = "Utah"
    VA = "Virginia"
    VI = "Virgin Islands"
    VT = "Vermont"
    WA = "Washington"
    WI = "Wisconson"
    WV = "West Virgina"
    WY = "Wyoming"

    def translate_state_code(code: str) -> str:
        """
        Given the state code, return full state name

        Args:
            code (str): 2 character state code

        Returns:
            str: Full state name
        """
        try:
            state_name = States[code].value
        except Exception:
            LOGGER.debug(f'Invalid state code: {code}')
            state_name = code

        return state_name

# =========================================================================================================    
@dataclass
class WeatherLocation():
    """
    Weather Location class

    Used in WeatherConditions.
    """
    latitude: float = 0.0
    longitude: float = 0.0
    location_name: str = None
    location_region: str = None

    def is_initialized(self) -> bool:
        """
        Is this weather location initialized.

        Returns:
            bool: True if lat/lon has been populated else False
        """
        return False if (self.latitude == 0.0 and self.longitude == 0.0) else True

